"""
Exceptions raised by the object clerk
"""


class ObjectClerkException(Exception):
    """
    Base Exception for the object clerk
    """


class ObjectNotFoundException(ObjectClerkException):
    """
    Exception when an object cannot be found either due to bucket or object key validity
    """


class ObjectVerificationException(ObjectClerkException):
    """
    Exception when checksums of source and destination data mismatch
    """


class ObjectSaveException(ObjectClerkException):
    """
    Exception when objects cannot be saved due to size or readability
    """



"""
Tests for the ObjectClerk API

Tests marked @pytest.mark.development require environment variables
to specify an s3 instance location and tokens
"""
from os import environ
from hashlib import md5
import boto3
import pytest

from object_clerk import ObjectClerk, ObjectNotFoundException

HOST = environ.get('HOST')
PORT = int(environ.get('PORT', 1))
ACCESS_KEY = environ.get('ACCESS_KEY')
SECRET_KEY = environ.get('SECRET_KEY')
# Configure retry to fail quickly on connection errors
RETRY = {'retry_delay': 1, 'retry_backoff': 1, 'retry_jitter': 1, 'retry_max_delay': 1, 'retry_tries': 1}
BUCKET = "test_object_clerk"


@pytest.fixture(scope="session")
def boto_client():
    endpoint_url = f"http://{HOST}:{PORT}/"
    return boto3.client(
            service_name="s3",
            aws_access_key_id=ACCESS_KEY,
            aws_secret_access_key=SECRET_KEY,
            endpoint_url=endpoint_url,
        )


@pytest.fixture(scope="module")
def object_store_data(boto_client):
    # Create Test Bucket
    boto_client.create_bucket(Bucket=BUCKET)
    with open("object_clerk/tests/example_object", mode='rb') as f:
        # Add Get file
        boto_client.upload_fileobj(f, Bucket=BUCKET, Key="get_file")
    with open("object_clerk/tests/example_object", mode='rb') as f:
        # Add Delete file
        boto_client.upload_fileobj(f, Bucket=BUCKET, Key="delete_file")
    with open("object_clerk/tests/example_object", mode='rb') as f:
        # Add Copy File
        boto_client.upload_fileobj(f, Bucket=BUCKET, Key="copy_file")
    yield
    # Delete objects
    boto_client.delete_object(Bucket=BUCKET, Key="get_file")
    boto_client.delete_object(Bucket=BUCKET, Key="delete_file")
    boto_client.delete_object(Bucket=BUCKET, Key="copy_file")
    # Delete Bucket
    boto_client.delete_bucket(Bucket=BUCKET)


@pytest.fixture()
def object_clerk():
    return ObjectClerk(
        host=HOST,
        port=PORT,
        access_key=ACCESS_KEY,
        secret_key=SECRET_KEY,
        **RETRY
    )


@pytest.mark.development
def test_get_object(object_clerk, object_store_data):
    response = object_clerk.get_object(BUCKET, "get_file")
    assert response == b'EXAMPLE OBJECT FOR TESTING OBJECT CLERK OPERATIONS'


@pytest.mark.development
@pytest.mark.parametrize(
    "bucket, key",
    [
        pytest.param(BUCKET, "foo", id="bad_key"),
        pytest.param("foo", "get_file", id="bad_bucket"),
        pytest.param("foo", "foo", id="bad_both"),
    ]
)
def test_get_object_not_found(object_clerk, object_store_data, bucket, key):
    with pytest.raises(ObjectNotFoundException):
        object_clerk.get_object(bucket, key)


@pytest.mark.development
def test_get_object_info(object_clerk, object_store_data):
    response = object_clerk.get_object_info(BUCKET, "get_file")
    with open("object_clerk/tests/example_object", mode='rb') as f:
        checksum = md5(f.read()).hexdigest()
    assert response['etag'][1:-1] == checksum


@pytest.mark.development
@pytest.mark.parametrize(
    "bucket, key",
    [
        pytest.param(BUCKET, "foo", id="bad_key"),
        pytest.param("foo", "get_file", id="bad_bucket"),
        pytest.param("foo", "foo", id="bad_both"),
    ]
)
def test_get_object_info_not_found(object_clerk, object_store_data, bucket, key):
    with pytest.raises(ObjectNotFoundException):
        object_clerk.get_object_info(bucket, key)


@pytest.mark.development
def test_copy_object(object_clerk, object_store_data):
    object_clerk.copy_object(BUCKET, "copy_file", BUCKET, "copied_file")
    source_checksum = object_clerk.get_object_info(BUCKET, "copy_file")['etag'][1:-1]
    destination_checksum = object_clerk.get_object_info(BUCKET, "copied_file")['etag'][1:-1]
    # clean up
    object_clerk.delete_object(BUCKET, "copied_file")
    assert source_checksum == destination_checksum


@pytest.mark.development
def test_delete_object(object_clerk, object_store_data):
    response = object_clerk.get_object_info(BUCKET, "delete_file")
    assert isinstance(response, dict)
    object_clerk.delete_object(BUCKET, "delete_file")
    with pytest.raises(ObjectNotFoundException):
        _ = object_clerk.get_object_info(BUCKET, "delete_file")


@pytest.mark.development
def test_upload_object(object_clerk, object_store_data):
    with open("object_clerk/tests/example_object", mode='rb') as f:
        object_clerk.upload_object(f, BUCKET, "upload_file")
    with open("object_clerk/tests/example_object", mode='rb') as f:
        checksum = md5(f.read()).hexdigest()
    etag = object_clerk.get_object_info(BUCKET, "upload_file")['etag']
    # clean up
    object_clerk.delete_object(BUCKET, "upload_file")
    assert etag[1:-1] == checksum


def test_pass():
    # work around for pipeline without any non development marked tests
    assert True



from hashlib import md5
import logging
from pathlib import Path
from sys import getsizeof
from typing import Union, Tuple
from io import BufferedReader, BytesIO
from boto3 import client as Boto3Client
from botocore.exceptions import ConnectionError, ClientError, HTTPClientError
from botocore.config import Config as BotoConfig
from retry import retry
from retry.api import retry_call

from object_clerk.exceptions import (
    ObjectClerkException,
    ObjectNotFoundException,
    ObjectSaveException,
    ObjectVerificationException,
)
from object_clerk.utils import mutate_client_exceptions

logger = logging.getLogger(__name__)  # TODO Logging pass

# Configure boto to not retry
BOTO_CONFIG = BotoConfig(connect_timeout=60, read_timeout=60, retries={"max_attempts": 0})

CONNECTION_RETRY_EXCEPTIONS = (ConnectionError, HTTPClientError)

INTERNAL_RETRY_CONFIG = {"tries": 3, "delay": 1}

# Set threshold at which the md5 checksum isn't valid to compare to the etag due to multipart uploads
MULTIPART_THRESHOLD = 1024 * 1024 * 1024 * 5  # 5 GB

__all__ = ["ObjectClerk"]


class ObjectClerk:
    """
    A wrapper for the following boto3 s3 client operations below the multipart upload threshold of 5 GB:
    - get_object : get_object
    - head_object : get_object_info
    - upload_fileobj : upload_object
    - copy_object : copy_object
    - delete_object : delete_object
    """
    def __init__(
        self,
        host: str,
        port: int,
        access_key: str,
        secret_key: str,
        retry_delay: int,
        retry_backoff: int,
        retry_jitter: Union[int, Tuple[int, int]],
        retry_max_delay: int,
        retry_tries: int = -1,
        use_ssl: bool = False,
    ):
        """
        Intialize the Object clerk with the location of the s3 gateway
        and configuration for retrying connection issues
        :param host: Host name or ip for an s3 Gateway
        :param port: Post the s3 gateway listens on
        :param access_key: Access Key for the gateway
        :param secret_key: Secret Key for the gateway
        :param retry_delay: initial delay between attempts for connection errors.
        :param retry_backoff: multiplier applied to delay between attempts to connect.
        :param retry_jitter: extra seconds added to delay between attempts to connect.
                   fixed if a number, random if a range tuple (min, max)
        :param retry_max_delay: the maximum value of delay between connection attempts.
        :param retry_tries: Number of time to retry connecting ot hte gateway. -1 for indefinite retries
        :param use_ssl: True for https and False for http
        """
        self.access_key = access_key
        self.secret_key = secret_key
        self.host = host
        self.port = port
        self.retry_delay = retry_delay
        self.retry_backoff = retry_backoff
        self.retry_jitter = retry_jitter
        self.retry_max_delay = retry_max_delay
        self.retry_tries = retry_tries
        self.use_ssl = use_ssl
        self.retry_config = {
            "tries": self.retry_tries,
            "delay": self.retry_delay,
            "max_delay": self.retry_max_delay,
            "backoff": self.retry_backoff,
            "jitter": self.retry_jitter,
            "exceptions": CONNECTION_RETRY_EXCEPTIONS,
        }
        protocol = "https" if self.use_ssl else "http"
        self.endpoint_url = f"{protocol}://{self.host}:{self.port}"
        self.client = Boto3Client(
            service_name="s3",
            aws_access_key_id=self.access_key,
            aws_secret_access_key=self.secret_key,
            endpoint_url=self.endpoint_url,
        )
        logger.info(str(self))

    def __repr__(self):
        return f"ObjectClerk(host={self.host}, port={self.port}, access_key={self.access_key}, " \
               f"secret_key=<SECRET KEY>, retry_delay={self.retry_delay}, retry_backoff={self.retry_backoff}, " \
               f"retry_jitter={self.retry_jitter}, retry_max_delay={self.retry_max_delay}, " \
               f"retry_tries={self.retry_tries}, use_ssl={self.use_ssl})"

    def __str__(self):
        return f"ObjectClerk connected to {self.endpoint_url} retrying connection errors with {self.retry_config}"

    @staticmethod
    def _get_object_checksum(object_stream: bytes) -> str:
        """

        :param object_stream:
        :return:
        """
        # Verify that the size doesn't exceed the limit at which the etag is != md5 checksum
        size = getsizeof(object_stream)
        if size > MULTIPART_THRESHOLD:
            logger.warning(f"Cannot verify the check sum of a file > {MULTIPART_THRESHOLD} bytes")
            raise ObjectClerkException(
                f"Cannot verify the check sum of a file > {MULTIPART_THRESHOLD} bytes"
            )
        return md5(object_stream).hexdigest()

    @staticmethod
    def _verify_checksum(object_checksum: str, etag: str) -> None:
        """

        :param object_checksum:
        :param etag:
        :return:
        """
        # Strip leading and trailing quotes on the etag if they exist
        etag = etag.replace('"', '')
        if object_checksum == etag:
            return

        logger.warning(
            f"Object checksum verification failed: check_sum={object_checksum}, etag={etag}"
        )
        raise ObjectVerificationException(
            f"Object checksum verification failed: check_sum={object_checksum}, etag={etag}"
        )

    @mutate_client_exceptions
    def _get_object(self, bucket: str, object_key: str) -> bytes:
        """

        :param bucket:
        :param object_key:
        :return:
        """
        logger.debug(f"Attempting object retrieval: bucket={bucket}, object_key={object_key}")
        return self.client.get_object(Bucket=bucket, Key=object_key)

    @retry(exceptions=ObjectVerificationException, **INTERNAL_RETRY_CONFIG)
    def get_object(
        self, bucket: str, object_key: str, verify_checksum: bool = True
    ) -> bytes:  # TODO check type
        """

        :param bucket:
        :param object_key:
        :param verify_checksum:
        :return:
        """
        response = retry_call(
            self._get_object,
            fkwargs={"bucket": bucket, "object_key": object_key},
            **self.retry_config,
        )
        object_stream = response.get("Body").read()
        etag = response.get("ETag")
        if verify_checksum:
            checksum = self._get_object_checksum(object_stream)
            self._verify_checksum(checksum, etag)
        return object_stream

    @mutate_client_exceptions
    def _get_object_info(self, bucket: str, object_key: str) -> dict:
        """

        :param bucket:
        :param object_key:
        :return:
        """
        logger.debug(f"Attempting object info retrieval: bucket={bucket}, object_key={object_key}")
        return self.client.head_object(Bucket=bucket, Key=object_key)

    @retry(exceptions=ObjectNotFoundException, **INTERNAL_RETRY_CONFIG)
    def get_object_info(self, bucket: str, object_key: str) -> dict:
        """

        :param bucket:
        :param object_key:
        :return:
        """
        response = retry_call(
            self._get_object_info,
            fkwargs={"bucket": bucket, "object_key": object_key},
            **self.retry_config,
        )
        return response["ResponseMetadata"].get("HTTPHeaders")

    @mutate_client_exceptions
    def _delete_object(self, bucket: str, object_key: str) -> None:
        """

        :param bucket:
        :param object_key:
        :return:
        """
        logger.debug(f"Attempting object delete: bucket={bucket}, object_key={object_key}")
        self.client.delete_object(Bucket=bucket, Key=object_key)

    def delete_object(self, bucket: str, object_key: str) -> None:
        """

        :param bucket:
        :param object_key:
        :return:
        """
        retry_call(
            self._delete_object,
            fkwargs={"bucket": bucket, "object_key": object_key},
            **self.retry_config,
        )

    @staticmethod
    def _to_readable_object(
        file: Union[str, Path, BufferedReader, BytesIO, bytes]
    ) -> BytesIO:
        """

        :param file:
        :return:
        """
        if isinstance(file, BytesIO):
            return file
        if isinstance(file, BufferedReader):
            return BytesIO(file.read())
        if isinstance(file, bytes):
            return BytesIO(file)
        if isinstance(file, str):
            file = Path(file)
        if not isinstance(file, Path):
            raise TypeError("file must be of one of type str, Path, BufferedReader, BytesIO, bytes")
        try:
            file = file.open(mode="rb")
        except OSError as e:
            raise ObjectSaveException(f"File cannot be opened: detail={e}")
        return BytesIO(file.read())

    @mutate_client_exceptions
    def _upload_object(self, file_obj: Union[BufferedReader, BytesIO], bucket: str, object_key: str) -> None:
        """

        :param file_obj:
        :param bucket:
        :param object_key:
        :return:
        """
        logger.debug(f"Attempting object upload: bucket={bucket}, object_key={object_key}")
        # Ensure byte stream is at the beginning for retries
        file_obj.seek(0)
        self.client.upload_fileobj(file_obj, Bucket=bucket, Key=object_key)

    @retry(exceptions=ObjectVerificationException, **INTERNAL_RETRY_CONFIG)
    def upload_object(
        self, file: Union[str, Path, BufferedReader], bucket: str, object_key: str, verify_checksum: bool = True
    ) -> None:
        """

        :param file:
        :param bucket:
        :param object_key:
        :param verify_checksum:
        :return:
        """
        file_obj = self._to_readable_object(file)
        # get checksum before boto operations
        checksum = self._get_object_checksum(file_obj.read())
        # reset back to start of the stream
        file_obj.seek(0)
        retry_call(
            self._upload_object,
            fkwargs={"file_obj": file_obj, "bucket": bucket, "object_key": object_key},
            **self.retry_config,
        )
        if verify_checksum:
            etag = self.get_object_info(bucket, object_key).get("etag")
            try:
                self._verify_checksum(checksum, etag)
            except ObjectVerificationException as e:
                logger.warning(f"Saved object does not match check sum: detail={e}")
                self.delete_object(bucket, object_key)
                logger.debug(f"Saved object is removed: bucket={bucket}, object_key={object_key}")
                raise e

    @mutate_client_exceptions
    def _copy_object(
        self, source_bucket: str, source_object_key: str, destination_bucket: str, destination_object_key: str
    ) -> None:
        """

        :param source_bucket:
        :param source_object_key:
        :param destination_bucket:
        :param destination_object_key:
        :return:
        """
        logger.debug(f"Attempting object retrieval: source_bucket={source_bucket}, "
                     f"source_object_key={source_object_key}, destination_bucket={destination_bucket}, "
                     f"destination_object_key={destination_object_key}")
        copy_source = {"Bucket": source_bucket, "Key": source_object_key}
        self.client.copy_object(
            Bucket=destination_bucket, Key=destination_object_key, CopySource=copy_source
        )

    @retry(exceptions=ObjectVerificationException, **INTERNAL_RETRY_CONFIG)
    def copy_object(
        self,
        source_bucket: str,
        source_object_key: str,
        destination_bucket: str,
        destination_object_key: str,
        verify_checksum: bool = True,
    ) -> None:
        """

        :param source_bucket:
        :param source_object_key:
        :param destination_bucket:
        :param destination_object_key:
        :param verify_checksum:
        :return:
        """
        source_object_info = self.get_object_info(source_bucket, source_object_key)
        if int(source_object_info["content-length"]) > MULTIPART_THRESHOLD:
            logger.warning(
                f"Attempt to copy file greater than multipart threshold: threshold={MULTIPART_THRESHOLD}, "
                f"size={source_object_info['content-length']}, bucket={source_bucket}, object_key={source_object_key}"
            )
            raise ObjectSaveException(
                f"API doesn't support uploads over {MULTIPART_THRESHOLD} bytes.  "
                f"Use the Boto3 mutltipart upload api."
            )

        retry_call(
            self._copy_object,
            fkwargs={
                "source_bucket": source_bucket,
                "source_object_key": source_object_key,
                "destination_bucket": destination_bucket,
                "destination_object_key": destination_object_key,
            },
            **self.retry_config,
        )

        if verify_checksum:
            destination_object_info = self.get_object_info(
                destination_bucket, destination_object_key
            )
            if source_object_info["etag"] == destination_object_info["etag"]:
                return
            self.delete_object(destination_bucket, destination_object_key)
            logger.warning(
                f"Copied object does not match check sum: source_bucket={source_bucket}, "
                f"source_object_key={source_bucket}, source_checksum={source_object_info['etag']}, "
                f"destination_bucket={destination_bucket}, destination_object_key={destination_object_key}, "
                f"destination_check_sum={destination_object_info['etag']}"
            )
            self.delete_object(destination_bucket, destination_object_key)
            logger.debug(
                f"Copied object is removed: bucket={destination_bucket}, object_key={destination_object_key}"
            )
            raise ObjectVerificationException(
                f"Object checksum verification failed: "
                f"source_etag={source_object_info['etag']}, "
                f"destination_etag={destination_object_info['etag']}"
            )
